//============================================================================
// CFAL12864G 128x64 OLED in SPI mode
// OLED_low.cpp : Lower-level OLED interface routines
// 2017 - 06 - 09 Brent A. Crosby
// Crystalfontz America, Inc. http://www.crystalfontz.com
//============================================================================
#include <Arduino.h>
#include <SPI.h>
// C:\Program Files (x86)\Arduino\hardware\arduino\avr\libraries\SPI\src\SPI.cpp
// C:\Program Files (x86)\Arduino\hardware\arduino\avr\libraries\SPI\src\SPI.h
#include "OLED_low.h"

//===========================================================================
//  The controller is a SSD1309. 
//  https://www.crystalfontz.com/controllers/Solomon%20Systech/SSD1309
#define SSD1309_00_SET_LOWER_COLUMN_ADDRESS_BIT       (0x00)
#define SSD1309_10_SET_UPPER_COLUMN_ADDRESS_BIT       (0x10)
#define SSD1309_20_MEMORY_ADDRESSING_MODE_PREFIX      (0x20)
#define SSD1309_00_ADRESSING_HORIZONTAL_PARAMETER     (0x00)
#define SSD1309_01_ADRESSING_VERTICAL_PARAMETER       (0x01)
#define SSD1309_02_ADRESSING_PAGE_PARAMETER           (0x02)
#define SSD1309_21_SET_COLUMN_ADDRESS_PREFIX          (0x21)
#define SSD1309_22_SET_PAGE_ADDRESS_PREFIX            (0x22)
#define SSD1309_26_SCROLL_SET_CONT_HORIZ_RIGHT_PREFIX (0x26)
#define SSD1309_27_SCROLL_SET_CONT_HORIZ_LEFT_PREFIX  (0x27)
#define SSD1309_29_SCROLL_SET_VERT_HORIZ_RIGHT_PREFIX (0x29)
#define SSD1309_2A_SCROLL_SET_VERT_HORIZ_LEFT_PREFIX  (0x2A)
#define SSD1309_2C_SCROLL_SET_ONE_HORIZ_RIGHT_PREFIX  (0x2C)
#define SSD1309_2D_SCROLL_SET_ONE_HORIZ_LEFT_PREFIX   (0x2D)
#define SSD1309_2E_SCROLL_DEACTIVATE                  (0x2E)
#define SSD1309_2F_SCROLL_ACTIVATE                    (0x2F)
#define SSD1309_40_SET_DISPLAY_START_LINE_BIT         (0x40)
#define SSD1309_81_CONTRAST_PREFIX                    (0x81)
#define SSD1309_A0_SEGMENT_REMAP_NORMAL               (0xA0)
#define SSD1309_A1_SEGMENT_REMAP_REVERSE              (0xA1)
#define SSD1309_A3_SCROLL_VERT_AREA_PREFIX            (0xA3)
#define SSD1309_A4_ENTIRE_DISPLAY_NORMAL              (0xA4)
#define SSD1309_A5_ENTIRE_DISPLAY_FORCE_ON            (0xA5)
#define SSD1309_A6_INVERSION_NORMAL                   (0xA6)
#define SSD1309_A7_INVERSION_INVERTED                 (0xA7)
#define SSD1309_A8_MULTIPLEX_RATIO_PREFIX             (0xA8)
#define SSD1309_AE_DISPLAY_OFF_SLEEP_YES              (0xAE)
#define SSD1309_AF_DISPLAY_ON_SLEEP_NO                (0xAF)
#define SSD1309_B0_SET_PAGE_START_ADDRESS_BIT         (0xB0)
#define SSD1309_C0_COM_DIRECTION_NORMAL               (0xC0)
#define SSD1309_C8_COM_DIRECTION_REVERSE              (0xC8)
#define SSD1309_D3_DISPLAY_VERT_OFFSET_PREFIX         (0xD3)
#define SSD1309_D5_CLOCK_DIVIDE_PREFIX                (0xD5)
#define SSD1309_D9_PRECHARGE_PERIOD_PREFIX            (0xD9)
#define SSD1309_DA_COM_PINS_CONFIGURATION_PREFIX      (0xDA)
#define SSD1309_DB_VCOMH_DESELECT_PREFIX              (0xDB)
#define SSD1309_DC_GPIO_CONFIGURATION_PREFIX          (0xDC)
#define SSD1309_E3_NOP                                (0xE3)
#define SSD1309_FD_LOCK_UNLOCK_PREFIX                 (0xFD)
#define SSD1309_16_LOCK_PARAMETER                     (0x16)
#define SSD1309_12_UNLOCK_PARAMETER                   (0x12)
//===========================================================================
void SPI_sendCommand(uint8_t command)
  {
  // Select the LCD's command register
  CLR_RS;
  // Select the LCD controller
  CLR_CS;
  //Send the command via SPI:
  SPI.transfer(command);
  // Deselect the LCD controller
  SET_CS;
  }
//============================================================================
void SPI_sendData(uint8_t data)
  {
  // Select the LCD's data register
  SET_RS;
  // Select the LCD controller
  CLR_CS;
  //Send the data via SPI:
  SPI.transfer(data);
  // Deselect the LCD controller
  SET_CS;
  }
//============================================================================
void Set_Address(uint8_t column, uint8_t page)
  {
  //Set column-lower nibble
  SPI_sendCommand(SSD1309_00_SET_LOWER_COLUMN_ADDRESS_BIT|(column&0x0F));
  //Set column-upper nibble
  SPI_sendCommand(SSD1309_10_SET_UPPER_COLUMN_ADDRESS_BIT|((column>>4)&0x0F));
  //Set page address, limiting from 0 to 7
  SPI_sendCommand(SSD1309_B0_SET_PAGE_START_ADDRESS_BIT|(page&0x07));
  }
//============================================================================
void Set_Brightness(uint8_t brightness)
  {
  //Set the "contrast" (brightness, max determined by IREF current)
  SPI_sendCommand(SSD1309_81_CONTRAST_PREFIX);
  SPI_sendCommand(brightness);
  }
//============================================================================
void Initialize_CFAL12864G(uint8_t brightness)
  {
  //Thump the hardware reset line.
  _delay_ms(1);
  CLR_RESET;
  _delay_ms(1);
  SET_RESET;
  _delay_ms(1);

  //Start with the display off (sleeping)
  SPI_sendCommand(SSD1309_AE_DISPLAY_OFF_SLEEP_YES);

  //Set the memory addressing mode to PAGE (increment column, no wrap)
  SPI_sendCommand(SSD1309_20_MEMORY_ADDRESSING_MODE_PREFIX);
  SPI_sendCommand(SSD1309_02_ADRESSING_PAGE_PARAMETER);

  //Point to the upper-left
  Set_Address(0,0);

  //Set the "contrast" (brightness, max determined by IREF current)
  SPI_sendCommand(SSD1309_81_CONTRAST_PREFIX);
  SPI_sendCommand(brightness);

  //Set start line to 0
  SPI_sendCommand(SSD1309_40_SET_DISPLAY_START_LINE_BIT);//Set Display Start Line

  //Set Segment remap for the CFAL12864G, we want the lower-left to be 0,0
  SPI_sendCommand(SSD1309_A1_SEGMENT_REMAP_REVERSE);

  //Ensure tha the "entire display force on", test mide is disabled. Read
  //the data from the RAM and display it as normal.
  SPI_sendCommand(SSD1309_A4_ENTIRE_DISPLAY_NORMAL);

  //Make sure that inversion is disabled.
  SPI_sendCommand(SSD1309_A6_INVERSION_NORMAL);

  //Set the multiplex ratio. 64 lines so 1/64
  SPI_sendCommand(SSD1309_A8_MULTIPLEX_RATIO_PREFIX);
  SPI_sendCommand(63);

  //Set COM directiom CFAL12864G, we want the lower-left to be 0,0
  SPI_sendCommand(SSD1309_C0_COM_DIRECTION_NORMAL);

  //No display vertical offset
  SPI_sendCommand(SSD1309_D3_DISPLAY_VERT_OFFSET_PREFIX);
  SPI_sendCommand(0);

  //Set clock frequency and division ratio
  SPI_sendCommand(SSD1309_D5_CLOCK_DIVIDE_PREFIX);
  //Fastest clock, smallest divisor.
  SPI_sendCommand(0xF0);
  // FFFF DDDD
  // |||| ||||-- DDDD = division, 0=>1, 1=>2 etc
  // ||||------- FFFF = frequency, 0 = slow 15 = fast

  //Set Pre-charge Period
  SPI_sendCommand(SSD1309_D9_PRECHARGE_PERIOD_PREFIX);
  SPI_sendCommand(0xF1);
  // 2222 1111
  // |||| ||||-- Phase 1 DCLK periods
  // ||||------- Phase 2 DCLK periods

  //Set COM Pins Hardware Configuration
  SPI_sendCommand(SSD1309_DA_COM_PINS_CONFIGURATION_PREFIX);
  // Alternate, Disable remap
  SPI_sendCommand(0x12);

  //Set Vcomh Deselect Level
  SPI_sendCommand(SSD1309_DB_VCOMH_DESELECT_PREFIX);
//// not in 1309 datasheet says it should be one of 00, 34 or 3C
  SPI_sendCommand(0x40);

  //Make sure the scroll is deactivated.
  SPI_sendCommand(SSD1309_2E_SCROLL_DEACTIVATE);

  //Turn the display on (wake)
  SPI_sendCommand(SSD1309_AF_DISPLAY_ON_SLEEP_NO);
  }
//============================================================================

