//===========================================================================
//
//  Code written for Seeeduino v4.2 runnung at 3.3v
//
//  CRYSTALFONTZ CFAL12864G 128x64 OLED in SPI mode
//
//  ref: https://www.crystalfontz.com/product/CFAL12864G
//
//  video: https://www.youtube.com/watch?v=PaO1ysq0seg
//
//  2017 - 06 - 09 Brent A. Crosby / Crystalfontz
//===========================================================================
//
//  !! IMPORTANT !!
//
//  The CFAL12864G is a 3.3v device. You need a 3.3v Arduino to operate this
//  code properly. We used a seeedunio v4.2 set to 3.3v:
//  http://www.seeedstudio.com/item_detail.html?p_id=2517
//
//  You need to supply +13v to pin 30 (Vcc) for the OLED drive.
//
//  IREF should be set to 10uA. Place a 1M Ohm resistor from pin 28 (IREF)
//  to ground (VSS).
//
//===========================================================================
//  This is free and unencumbered software released into the public domain.
//
//  Anyone is free to copy, modify, publish, use, compile, sell, or
//  distribute this software, either in source code form or as a compiled
//  binary, for any purpose, commercial or non-commercial, and by any
//  means.
//
//  In jurisdictions that recognize copyright laws, the author or authors
//  of this software dedicate any and all copyright interest in the
//  software to the public domain. We make this dedication for the benefit
//  of the public at large and to the detriment of our heirs and
//  successors. We intend this dedication to be an overt act of
//  relinquishment in perpetuity of all present and future rights to this
//  software under copyright law.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//  OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//  OTHER DEALINGS IN THE SOFTWARE.
//
//  For more information, please refer to <http://unlicense.org/>
//===========================================================================
#include <SPI.h>
// C:\Program Files (x86)\Arduino\hardware\arduino\avr\libraries\SPI\src\SPI.cpp
// C:\Program Files (x86)\Arduino\hardware\arduino\avr\libraries\SPI\src\SPI.h

// There is limited memory on the Arduino, so we can have the SD card or
// we could have a RAM display buffer, but not both.

#include <SD.h>
// C:\Program Files (x86)\Arduino\libraries\SD\src\SD.cpp
// C:\Program Files (x86)\Arduino\libraries\SD\src\SD.h
//============================================================================
// LCD SPI & control lines
//   ARD      | Port | LCD
// -----------+------+-------------------------
//  #7/D7     |  PD7 | SD_CS
//  #8/D8     |  PB0 | LCD_RS
//  #9/D9     |  PB1 | LCD_RESET
// #10/D10    |  PB2 | LCD_CS_NOT (or SPI SS)
// #11/D11    |  PB3 | LCD_MOSI   (hardware SPI)
// #12/D12    |  PB4 | not used   (would be MISO)
// #13/D13    |  PB5 | LCD_SCK    (hardware SPI)

#define CLR_RS    (PORTB &= ~(0x01))
#define SET_RS    (PORTB |=  (0x01))
#define CLR_RESET (PORTB &= ~(0x02))
#define SET_RESET (PORTB |=  (0x02))
#define CLR_CS    (PORTB &= ~(0x04))
#define SET_CS    (PORTB |=  (0x04))
#define CLR_MOSI  (PORTB &= ~(0x08))
#define SET_MOSI  (PORTB |=  (0x08))
#define CLR_SCK   (PORTB &= ~(0x20))
#define SET_SCK   (PORTB |=  (0x20))

#define SD_CS (A0)

//===========================================================================
//  The controller is a SSD1309. 
//  https://www.crystalfontz.com/controllers/Solomon%20Systech/SSD1309
#define SSD1309_00_SET_LOWER_COLUMN_ADDRESS_BIT       (0x00)
#define SSD1309_10_SET_UPPER_COLUMN_ADDRESS_BIT       (0x10)
#define SSD1309_20_MEMORY_ADDRESSING_MODE_PREFIX      (0x20)
#define SSD1309_00_ADRESSING_HORIZONTAL_PARAMETER     (0x00)
#define SSD1309_01_ADRESSING_VERTICAL_PARAMETER       (0x01)
#define SSD1309_02_ADRESSING_PAGE_PARAMETER           (0x02)
#define SSD1309_21_SET_COLUMN_ADDRESS_PREFIX          (0x21)
#define SSD1309_22_SET_PAGE_ADDRESS_PREFIX            (0x22)
#define SSD1309_26_SCROLL_SET_CONT_HORIZ_RIGHT_PREFIX (0x26)
#define SSD1309_27_SCROLL_SET_CONT_HORIZ_LEFT_PREFIX  (0x27)
#define SSD1309_29_SCROLL_SET_VERT_HORIZ_RIGHT_PREFIX (0x29)
#define SSD1309_2A_SCROLL_SET_VERT_HORIZ_LEFT_PREFIX  (0x2A)
#define SSD1309_2C_SCROLL_SET_ONE_HORIZ_RIGHT_PREFIX  (0x2C)
#define SSD1309_2D_SCROLL_SET_ONE_HORIZ_LEFT_PREFIX   (0x2D)
#define SSD1309_2E_SCROLL_DEACTIVATE                  (0x2E)
#define SSD1309_2F_SCROLL_ACTIVATE                    (0x2F)
#define SSD1309_40_SET_DISPLAY_START_LINE_BIT         (0x40)
#define SSD1309_81_CONTRAST_PREFIX                    (0x81)
#define SSD1309_A0_SEGMENT_REMAP_NORMAL               (0xA0)
#define SSD1309_A1_SEGMENT_REMAP_REVERSE              (0xA1)
#define SSD1309_A3_SCROLL_VERT_AREA_PREFIX            (0xA3)
#define SSD1309_A4_ENTIRE_DISPLAY_NORMAL              (0xA4)
#define SSD1309_A5_ENTIRE_DISPLAY_FORCE_ON            (0xA5)
#define SSD1309_A6_INVERSION_NORMAL                   (0xA6)
#define SSD1309_A7_INVERSION_INVERTED                 (0xA7)
#define SSD1309_A8_MULTIPLEX_RATIO_PREFIX             (0xA8)
#define SSD1309_AE_DISPLAY_OFF_SLEEP_YES              (0xAE)
#define SSD1309_AF_DISPLAY_ON_SLEEP_NO                (0xAF)
#define SSD1309_B0_SET_PAGE_START_ADDRESS_BIT         (0xB0)
#define SSD1309_C0_COM_DIRECTION_NORMAL               (0xC0)
#define SSD1309_C8_COM_DIRECTION_REVERSE              (0xC8)
#define SSD1309_D3_DISPLAY_VERT_OFFSET_PREFIX         (0xD3)
#define SSD1309_D5_CLOCK_DIVIDE_PREFIX                (0xD5)
#define SSD1309_D9_PRECHARGE_PERIOD_PREFIX            (0xD9)
#define SSD1309_DA_COM_PINS_CONFIGURATION_PREFIX      (0xDA)
#define SSD1309_DB_VCOMH_DESELECT_PREFIX              (0xDB)
#define SSD1309_DC_GPIO_CONFIGURATION_PREFIX          (0xDC)
#define SSD1309_E3_NOP                                (0xE3)
#define SSD1309_FD_LOCK_UNLOCK_PREFIX                 (0xFD)
#define SSD1309_16_LOCK_PARAMETER                     (0x16)
#define SSD1309_12_UNLOCK_PARAMETER                   (0x12)
//===========================================================================
void SPI_sendCommand(uint8_t command)
  {
  // Select the LCD's command register
  CLR_RS;
  // Select the LCD controller
  CLR_CS;
  //Send the command via SPI:
  SPI.transfer(command);
  // Deselect the LCD controller
  SET_CS;
  }
//============================================================================
void SPI_sendData(uint8_t data)
  {
  // Select the LCD's data register
  SET_RS;
  // Select the LCD controller
  CLR_CS;
  //Send the data via SPI:
  SPI.transfer(data);
  // Deselect the LCD controller
  SET_CS;
  }
//============================================================================
void Set_Address(uint8_t column, uint8_t page)
  {
  //Set column-lower nibble
  SPI_sendCommand(SSD1309_00_SET_LOWER_COLUMN_ADDRESS_BIT|(column&0x0F));
  //Set column-upper nibble
  SPI_sendCommand(SSD1309_10_SET_UPPER_COLUMN_ADDRESS_BIT|((column>>4)&0x0F));
  //Set page address, limiting from 0 to 7
  SPI_sendCommand(SSD1309_B0_SET_PAGE_START_ADDRESS_BIT|(page&0x07));
  }
//============================================================================
void Set_Brightness(uint8_t brightness)
  {
  //Set the "contrast" (brightness, max determined by IREF current)
  SPI_sendCommand(SSD1309_81_CONTRAST_PREFIX);
  SPI_sendCommand(brightness);
  }
//============================================================================
void Initialize_CFAL12864G(uint8_t brightness)
  {
  //Thump the hardware reset line.
  _delay_ms(1);
  CLR_RESET;
  _delay_ms(1);
  SET_RESET;
  _delay_ms(1);

  //Start with the display off (sleeping)
  SPI_sendCommand(SSD1309_AE_DISPLAY_OFF_SLEEP_YES);

  //Set the memory addressing mode to PAGE (increment column, no wrap)
  SPI_sendCommand(SSD1309_20_MEMORY_ADDRESSING_MODE_PREFIX);
  SPI_sendCommand(SSD1309_02_ADRESSING_PAGE_PARAMETER);

  //Point to the upper-left
  Set_Address(0,0);

  //Set the "contrast" (brightness, max determined by IREF current)
  SPI_sendCommand(SSD1309_81_CONTRAST_PREFIX);
  SPI_sendCommand(brightness);

  //Set start line to 0
  SPI_sendCommand(SSD1309_40_SET_DISPLAY_START_LINE_BIT);//Set Display Start Line

  //Set Segment remap for the CFAL12864G, we want the lower-left to be 0,0
  SPI_sendCommand(SSD1309_A1_SEGMENT_REMAP_REVERSE);

  //Ensure tha the "entire display force on", test mide is disabled. Read
  //the data from the RAM and display it as normal.
  SPI_sendCommand(SSD1309_A4_ENTIRE_DISPLAY_NORMAL);

  //Make sure that inversion is disabled.
  SPI_sendCommand(SSD1309_A6_INVERSION_NORMAL);

  //Set the multiplex ratio. 64 lines so 1/64
  SPI_sendCommand(SSD1309_A8_MULTIPLEX_RATIO_PREFIX);
  SPI_sendCommand(63);

  //Set COM directiom CFAL12864G, we want the lower-left to be 0,0
  SPI_sendCommand(SSD1309_C0_COM_DIRECTION_NORMAL);

  //No display vertical offset
  SPI_sendCommand(SSD1309_D3_DISPLAY_VERT_OFFSET_PREFIX);
  SPI_sendCommand(0);

  //Set clock frequency and division ratio
  SPI_sendCommand(SSD1309_D5_CLOCK_DIVIDE_PREFIX);
  //Fastest clock, smallest divisor.
  SPI_sendCommand(0xF0);
  // FFFF DDDD
  // |||| ||||-- DDDD = division, 0=>1, 1=>2 etc
  // ||||------- FFFF = frequency, 0 = slow 15 = fast

  //Set Pre-charge Period
  SPI_sendCommand(SSD1309_D9_PRECHARGE_PERIOD_PREFIX);
  SPI_sendCommand(0xF1);
  // 2222 1111
  // |||| ||||-- Phase 1 DCLK periods
  // ||||------- Phase 2 DCLK periods

  //Set COM Pins Hardware Configuration
  SPI_sendCommand(SSD1309_DA_COM_PINS_CONFIGURATION_PREFIX);
  // Alternate, Disable remap
  SPI_sendCommand(0x12);

  //Set Vcomh Deselect Level
  SPI_sendCommand(SSD1309_DB_VCOMH_DESELECT_PREFIX);
//// not in 1309 datasheet says it should be one of 00, 34 or 3C
  SPI_sendCommand(0x40);

  //Make sure the scroll is deactivated.
  SPI_sendCommand(SSD1309_2E_SCROLL_DEACTIVATE);

  //Turn the display on (wake)
  SPI_sendCommand(SSD1309_AF_DISPLAY_ON_SLEEP_NO);
  }
//----------------------------------------------------------------------------
void Fill_Display(uint8_t data1, uint8_t data2)
  {
  uint8_t
    page;
  uint8_t
    column;
  for(page=0;page<8;page++)
    {
    //Set the LCD to the left of this line.
    Set_Address(0,page);
    //Move across the columns, alternating the two data
    //bytes.
    for(column=0;column<64;column++)
      {
      SPI_sendData(data1);
      SPI_sendData(data2);
      }
    }              
  }
//----------------------------------------------------------------------------
void show_BMPs_in_root(void)
  {
  File
    root_dir;
  root_dir = SD.open("/");
  if(0 == root_dir)
    {
    Serial.println(F("show_BMPs_in_root(): Can't open \"root\""));
    return;
    }
  File
    bmp_file;

  while(1)
    {
    bmp_file = root_dir.openNextFile();
    if (0 == bmp_file)
      {
      // no more files, break out of while()
      // root_dir will be closed below.
      break;
      }
    //Skip directories (what about volume name?)
    if(0 == bmp_file.isDirectory())
      {
      //The file name must include ".BMP"
      if(0 != strstr(bmp_file.name(),".BMP"))
        {
        //The BMP must be exactly 1088 long
        //(this is correct for 128x64 1-bit)
        if(1088 == bmp_file.size())
          {
          Serial.print(F("Display BMP: "));
          Serial.println(bmp_file.name());
          //Jump over BMP header. BMP must be 128x64 1-bit
          bmp_file.seek(0x3E);
    
          //Since we are limited in memory, break the line up from
          // 240*3 = 720 bytes into three chunks of 80 pixels
          // each 80*3 = 240 bytes.
          //Making this static speeds it up slightly (10ms)
          //Reduces flash size by 114 bytes, and uses 128 bytes.
          static uint8_t
            line_buffer[128];
          for(uint8_t line=0;line<8;line++)
            {
            //Set the LCD to the left of this line.
            Set_Address(0,line);
            
            //Read 8 lines out of the BMP
            bmp_file.read(line_buffer,128);

            //BMPs store data
            //lowest line first -- bottom up.
            //Set_LCD_for_write_at_X_Y(0,line);

            uint8_t
              h_mask;
            h_mask=0x80;
            for(uint8_t col=0;col<128;col++)
              {
              uint8_t
                sub_line;
              uint8_t
                v_mask;
              v_mask=0x01;
              uint8_t
                this_v_byte;
              this_v_byte=0;
              for(sub_line=0;sub_line<8;sub_line++)
                {
                if(line_buffer[(sub_line<<4)+(col>>3)] & h_mask)
                  {
                  this_v_byte|=v_mask;
                  }

                v_mask<<=1;
                }

              SPI_sendData(this_v_byte);
              //delay(20);
              
              h_mask>>=1;
              if(0x00 == h_mask)
                h_mask=0x80;
              }
            }
          }
        }
      }
    //Release the BMP file handle
    bmp_file.close();
    delay(2000);
    }
  //Release the root directory file handle
  root_dir.close();
  }
//============================================================================
void setup( void )
  {
  // LCD SPI & control lines
  //   ARD      | Port | LCD
  // -----------+------+-------------------------
  //  #7/D7     |  PD7 | SD_CS    
  //  #8/D8     |  PB0 | LCD_RS
  //  #9/D9     |  PB1 | LCD_RESET
  // #10/D10    |  PB2 | LCD_CS_NOT (or SPI SS)
  // #11/D11    |  PB3 | LCD_MOSI   (hardware SPI)
  // #12/D12    |  PB4 | not used   (would be MISO)
  // #13/D13    |  PB5 | LCD_SCK    (hardware SPI)
  // #23/D14/A0 |  PC0 | Touch XL   (only used on TS modules)
  // #24/D15/A1 |  PC1 | Touch XR   (only used on TS modules)
  // #25/D16/A2 |  PC2 | Touch YD   (only used on TS modules)
  // #26/D17/A3 |  PC3 | Touch YU   (only used on TS modules)
  DDRB |= 0x2F;

  //Drive the ports to a reasonable starting state.
  CLR_RESET;
  CLR_RS;
  SET_CS;
  CLR_MOSI;
  CLR_SCK;

  //Set the SD card CS to output, not selected.
  pinMode(SD_CS, OUTPUT);
  digitalWrite(SD_CS, HIGH);

  //debug console
  Serial.begin(9600);
  Serial.println(F("setup()"));

  // Initialize SPI. By default the clock is 4MHz.
  SPI.begin();
  //Bump the clock to 8MHz. Appears to be the maximum.
  SPI.beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE0));

  //Fire up the SPI OLED
  Serial.println(F("Initialize_CFAL12864G()"));
  Initialize_CFAL12864G(255);

  //Push out a few filles screens
  //Checkerboard odd
  Fill_Display(0xAA,0x55);
  delay(500);
  //Checkerboard even
  Fill_Display(0x55,0xAA);
  delay(500);
  //Horizontal lines odd
  Fill_Display(0xAA,0xAA);
  delay(500);
  //Horizontal lines even
  Fill_Display(0x55,0x55);
  delay(500);
  //Vertical lines odd
  Fill_Display(0xFF,0x00);
  delay(500);
  //Vertical lines even
  Fill_Display(0x00,0xFF);
  delay(500);
  //All on
  Fill_Display(0xFF,0xFF);
  delay(500);
  //All off
  Fill_Display(0x00,0x00);
  delay(500);
 
  // For the Seeduino I am using, the default speed of SPI_HALF_SPEED
  // set in C:\Program Files (x86)\Arduino\libraries\SD\src\SD.cpp
  // results in a 4MHz clock.
  //
  // If you change this function call in SDClass::begin() of SD.cpp
  // from:
  //
  //  return card.init(SPI_HALF_SPEED, csPin) &&
  //         volume.init(card) &&
  //         root.openRoot(volume);
  //
  // to:
  //
  //  return card.init(SPI_FULL_SPEED, csPin) &&
  //         volume.init(card) &&
  //         root.openRoot(volume);
  //
  // That appears to make the SD library talk at 8MHz.
  //
  // Alas, it only improves a full-screen update from a dismally slow
  // time of 1.81 seconds to the nearly hopeless time of 1.35 seconds.
  //
  if(!SD.begin(SD_CS)) 
    {
    Serial.println(F("uSD Card failed to initialize, or not present"));
    //Reset the SPI clock to fast. SD card library does not clean up well.
    //Bump the clock to 8MHz. Appears to be the maximum.
    SPI.beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE0));    
    while(1);
    }
  else
    {
    Serial.println(F("uSD Card initialized."));
    }
  }
//============================================================================
void loop(void)
  {
  Serial.println(F("loop()"));
  show_BMPs_in_root();    
  }
//============================================================================

